// =============================================================================
//
// Introduction to Dafny, the programming language
//
// =============================================================================
//
// At its core Dafny is an imperative programming language.
//
// This file demonstrates that that means, and we will do this by
// example by look at operating over 'lists'.
//
// We will provide some sample code, and method prototypes.
//
// We will answer some questions together, but we will give you this
// file unmodified to try outside of the lecture.
//
// A lengthier look at Dafny is on MyPlace.
//

// =============================================================================
//
// Let us begin with a method that returns a sequence of Integers.
//
// =============================================================================

method MyFirstList()
  returns (res : seq<int>)
{
  return [1,2,3,4];
}


/* [ Marker ] ==================================================================

// [ Question ]
//
// Can you write a method that returns the list `[1,2,3,4,5,6,7,8]`?



// [ Question ]
//
// Can you finish this method that checks if an integer is positive?
// It will need to return Bool.
method IsPositive(i : int) returns (res : bool)
{
  return false;
}

method {:test} TestIsPositive()
{
  var res := IsPositive(1);
  expect res == true;

  res := IsPositive(-1);
  expect res == false;

  res := IsPositive(0);
  expect res == true;

}

// =============================================================================
//
// Questions on writing methods that operate on lists and integers.
//
// =============================================================================


// [ Question ]
//
// Can you finish this method that *should* append two lists together?

method MyAppend<T>(xs : seq<T>, ys : seq<T>)
  returns (res : seq<T>)
{
  return [];
}


method {:test} TestMyAppend()
{

  var xs : seq<int> := [1,2,3,4];
  var ys : seq<int> := [5,6,7,8];

  var zs := MyAppend(xs,ys);

  expect zs == [1,2,3,4,5,6,7,8];
}

// [ Question ]
//
// Can you finish this method that returns the first element (the head) of a list?
//
// Hint: `Maybe`, is defined later on in the file.

method MyHead<T>(xs : seq<T>)
  returns (res : Maybe<T>)
{
  return Nothing;
}


method {:test} TestMyHead()
{

  var xs : seq<int> := [1,2,3,4];

  var zs := MyHead(xs);

  expect zs == Just(1);
}

// [ Question ]
//
// Can you finish this method that returns all but the first element of a list
// (the tail)?

method MyTail<T>(xs : seq<T>)
  returns (res : seq<T>)
{
  return [];
}


method {:test} TestMyTail()
{

  var xs : seq<int> := [1,2,3,4];

  var zs := MyTail(xs);

  expect zs == [2,3,4];
}

// =============================================================================
//
// Questions on printing things about lists to STDOUT.
//
// =============================================================================

// [ NOTE ] These tests will pass, but show the wrong result.


// [ Question ]
//
// Can you finish this method that prints out each element in the sequence?


method MyShowList<T>(xs : seq<T>)
{
  print "Nothing", "\n";
}


method {:test} TestMyShowList()
{

  var xs := MyFirstList(); // [1,2,3,4]

  MyShowList(xs);

  xs := [];

  MyShowList(xs);
}

// [ Question ]
//
// Can you finish this method that prints out each element in the sequence,
// and if the sequence is empty, then prints "Empty List"?


method MyShowList2<T>(xs : seq<T>)
{
  print "Nothing", "\n";
}


method {:test} TestMyShowList2()
{

  var xs : seq<int> := [1,2,3,4];

  MyShowList2(xs);

  xs := [];

  MyShowList2(xs);
}


// =============================================================================
//
// Questions on datatypes
//
// =============================================================================


// Here is a datatype to represent optional things.
//
// Given a value of type `Maybe<T>`, this value will either be:
//
// + `Nothing`, if there is no value;
// + `Just(t)`, if there is a value `t`.

datatype Maybe<T> = Nothing | Just(t : T)

// [ Question ]
//
// Can you finish this method that returns the following string?
//
// + "Is Nothing" if `m` is `Nothing`;
// + "Is a Value" if `m` is `Just(t)`;


method MyMaybeToString<T>(m : Maybe<T>)
  returns (res : string)
{
  res := "";
  print "Nothing", "\n";
}


method {:test} TestMyMaybeToString()
{

  var x : Maybe<string> := Nothing;
  var res := MyMaybeToString(x);

  expect res == "Is Nothing";

  res := MyMaybeToString(Just(1));

  expect res == "Is a Value";

}


== [ Marker ] =============================================================== */


// -- [ EOF ]

// =============================================================================
//
// Header
//
// ========================================================================== */

// [ NOTE ]
//
//
