||| Here we look at Peano and `plusCheck`.
module PlusCheck

-- We need Decidable Equality to tell us if two peano numbers are equal.
import Decidable.Equality

data Peano = Zero      -- There is nothing, the base case;
                       -- or
           | Inc Peano -- there is one more.


data Plus : (a,b,c : Peano) -> Type
  where
    ||| Base case.
    |||
    ||| ```
    |||    0 + b = b
    ||| ```
    ZI : Plus Zero b b

    ||| Induction on left operand.
    |||
    ||| ```
    ||| S(a) + b = S(a+b)
    ||| ```
    IA : (step : Plus      x  b      c)
              -> Plus (Inc x) b (Inc c)


-- Instances on Uninhabited  to help us along.

Uninhabited (Plus Zero (Inc x) Zero) where
  uninhabited ZI impossible

Uninhabited (Plus Zero Zero (Inc x)) where
  uninhabited ZI impossible

Uninhabited (Plus (Inc x) b Zero) where
  uninhabited ZI impossible


Uninhabited (Zero = Inc x) where
  uninhabited Refl impossible

||| Instance of decidable equality for example and to hep us with `plusCheck`
DecEq Peano where
  decEq Zero Zero = Yes Refl
  decEq Zero (Inc x)
    = No absurd
  decEq (Inc x) Zero
    = No (negEqSym absurd)
  decEq (Inc x) (Inc y) with (decEq x y)
    decEq (Inc x) (Inc y) | (Yes prf) with (prf)
      decEq (Inc y) (Inc y) | (Yes Refl) | Refl
      = Yes Refl
    decEq (Inc x) (Inc y) | (No contra)
      = No (\case Refl => contra Refl)


plusCheck : (a,b,c : Peano) -> Dec (Plus a b c)
-- Base Case
plusCheck Zero Zero Zero
  = Yes ZI

-- Some Failing cases
plusCheck Zero Zero (Inc x)
  = No absurd

plusCheck Zero (Inc x) Zero
  = No absurd

-- Cheating a little and 'proving' that the `x` and `y` are different.
plusCheck Zero (Inc x) (Inc y) with (decEq x y)
  plusCheck Zero (Inc x) (Inc y) | (Yes prf) with (prf)
    plusCheck Zero (Inc y) (Inc y) | (Yes prf) | Refl
      = Yes ZI
  plusCheck Zero (Inc x) (Inc y) | (No contra)
    = No (\case ZI => contra Refl)
    -- ^ this is a proof that: If we know that `x` and `y` are
    -- different but we think they are the same then we have falsity.

plusCheck (Inc x) b Zero
  = No absurd

plusCheck (Inc x) b (Inc y) with (plusCheck x b y)
  plusCheck (Inc x) b (Inc y) | (Yes prf) = Yes (IA prf)
  plusCheck (Inc x) b (Inc y) | (No contra)
    = No (\case (IA step) => contra step)
    -- ^ if we know that the inductive case is false then we have proof of void.


-- [ EOF ]
