/* CS886 Week 4 -- Bounds Checking using Dafny

   ## What we will cover

   In this part of the Live Coding session we will cover approaches to bounds checking in Dafny.

   Dafny has support for various 'list' style data structures:

   + `seq<t>` Sequences, unbounded linked lists, can contain repeated elements;
   + `set<t>` sets, containing unbounded unique elements;
   + `array<t>`, bounded lists

 */

/* [ Marker ] ==================================================================

/* =============================================================================

   Examples of unsafe indexing

   ========================================================================== */

method ArrayIdxUnsafe<T>(xs : array<T>, idx : int) returns (x : T)
{
  return xs[idx];
}

method SeqIdxUnsafe<T>(xs : seq<T>, idx : int) returns (x : T)
{
  return xs[idx];
}

/* =============================================================================

   Examples of safe indexing

   ========================================================================== */

method SeqIdxSafe<T>(xs : seq<T>, idx : int)
  returns (x : T)
  requires 0 <= idx < |xs|
{
  return xs[idx];
}

method SeqIdxSafeUsage<T>(xs : seq<T>, idx : int)
  returns (x : T)
{
  var res := SeqIdxSafe(xs,idx);
  // ^ fails here as precondition for `SeqIdxSafe` cannot be proven!
  return res;
}

method SeqIdxSafeUsage2<T>(xs : seq<T>, idx : int)
  returns (x : T)
{
  // How to fix?
  var res := SeqIdxSafe(xs,idx);
  return res;
}


/* =============================================================================

   A Safe Wrapper

   ========================================================================== */

/* There are no private fields in Dafny.

   To create safe wrappers, we need to hide the field being wrapper.
   We can do this using modules and expore sets.

*/

module SafeList
{
  // This controls what methods can be seen, and what can be used in
  // specifications.
  export reveals SafeList provides SafeList.Add, SafeList.Del

  // This will fail. Why?
  class SafeList<T>
  {
    var xs : seq<T>

    constructor() { this.xs := []; }

    method Add(x : T)
      modifies this
      ensures |xs| == old(|xs|) + 1
    {
      this.xs := [x] + this.xs;
    }

    method Del(x : T)
      modifies this
      ensures |xs| == old(|xs|) - 1
    {
      this.xs := this.xs[1..];
    }
  }

}


   [ Marker ] =============================================================== */

// -- [ EOF ]
