/* CS886 Week 4 -- Fake Substructural Typing.

   ## What we will cover

   In this part of the Live Coding session we will cover approaches to faking substructural typing to restrict 'variable' usage in Dafny.

   Our approach will not necessarily be sound, nor complete.

 */

/* =============================================================================

   Reminder

   ========================================================================== */

method PopulatingPair<A>(x : A, y : A) returns (p : (A,A))
{
  return (x,x);
}

/* =============================================================================

   Faking linear variables

   ========================================================================== */

module Linear
{
  // This controls what methods can be seen, and what can be used in
  // specifications.
  export reveals Linear provides Linear.Get, Linear.Get2, Linear.free
  class Linear<T>
  {
    var x : T

    ghost var free : bool

    constructor(x : T)
    {
      this.x := x;
      this.free := true;
    }

    method Get() returns (x : T)
      modifies this
      requires this.free
      ensures old(this.free) != free
    {
      this.free := false;
      return this.x;
    }

    method Get2() returns (x : T)
      modifies this
    {
      this.free := true;
      return this.x;
    }
  }
}

import opened Linear

  // An Example
/*
method PopulatingPair2<A>(x : Linear<A>, y : Linear<A>) returns (p : (A,A))
{
  var a := x.Get();
  var b := x.Get();
  return (a,a);
}


*/

/* [ NOTE ]

   + Fails because we do not know if `x` or `y` are free!

*/

method PopulatingPair3<A>(x : Linear<A>, y : Linear<A>)
  returns (p : (A,A))
  modifies x
  modifies y
//  requires x.free
  requires y.free
//  ensures  !x.free
  ensures  !y.free
{
  var a := x.Get2();
//  var b := x.Get();
  var b := y.Get();
  return (a,a);
}


/* [ Marker ] ==================================================================



   [ Marker ] =============================================================== */

// -- [ EOF ]
