||| Here we demonstrate some of Idris2 features relating to QTT
module Substructural


import Data.List

{- =============================================================================

   QTT in Idris 2

-}

pairBad : (x,y : a) -> (a, a)
pairBad x y = (x,x)
            -- ^ ^ --

pairGood : (1 x : a) -> (1 y : a) -> (a, a)

{- [ NOTES ]

1. Show definition of Pair and LPair and Erased Pair
2. talk about linear contexts

-}

{- [ Marker ] ==================================================================






































{- =============================================================================

   QTT in Idris 2

-}

||| Showing Head using implicit function arguments
head : (xs  : List a)
    -> (prf : NonEmpty xs)
    => a
  -- ^ type class constraint, interface constraint
  -- Idris desugars interface constraints to a data type
  -- Idris Interfaces and their instances are just data.
head (x :: xs) {prf = IsNonEmpty} = x


||| Showing `head` but with quantifies explicit
head' : {0 a   : Type}
     -> (xs    : List a)
     -> (0 prf : NonEmpty xs)
              => a
head' (x :: xs) {prf = IsNonEmpty} = x

-- What happens if we do not have proof search?

||| Showing `head` but sans proof search
head'' : {0 a   : Type}
      -> (xs    : List a)
      -> (0 prf : NonEmpty xs)
                -> a
head'' xs prf = head xs {prf = prf}

||| Showing `head` but with proofs used at compile time.
head''' : {0 a   : Type}
       -> (xs    : List a)
       -> (prf   : NonEmpty xs)
                -> a






















{- =============================================================================

   Using Linearity to ensure data is used.

-}

||| A API to work with file handles.
|||
||| Within Idris2 the File API is embedded within the `IO` computation
||| context.
|||
||| This is somewhat illustrative, and not safe.
|||
||| We are missing some behaviours:
|||
||| + file modes
||| + safe return values
namespace Files

  ||| An abstract datatype to represent File handles or some externally
  ||| allocated memory.
  data FH : Type where


  interface File where
    openFile : String -> ((s : FH) -> IO ()) -> IO ()

    readLine : FH -> IO String

    writeLine : FH -> String -> IO ()

    closeFile : FH -> IO ()

  example : File => IO ()
  example
    = openFile "Hello.txt"
      $ \fh =>
        do  str <- readLine fh
            writeLine fh "Write"
            closeFile fh
            writeLine fh "Write"


  namespace Linear

    interface FileL where
      openFile :  (fname  : String)
               -> (1 cont : (1 fh : FH) -> IO ())
               -> IO ()

      closeFile : (1 fh : FH) -> IO ()

      readLine : (1 fh   : FH)
              -> (1 cont : (1 res : LPair FH String) -> IO ())
                        -> IO ()

      writeLine : (1 fh   : FH)
               -> (1 str  : String)
               -> (1 cont : (1 res : LPair FH ()) -> IO ())
                         -> IO ()



== [ Marker ] =============================================================== -}

-- [ EOF ]
