module Common
{
  module Result
  {
    datatype Result<E,T> = Error(err : E) | Success(t : T)

    predicate isSuccess<E,T>(m : Result<E,T>)
    {
      match m
      {
      case Error(_) => false
      case Success(_) => true
      }
    }

    function result<E,T,A>(err : E -> A, res : T -> A, r : Result<E,T>) : A
    {
      match r
      {
      case Error(e) => err(e)
      case Success(s) => res(s)
      }
    }

    predicate applySuccess<E,T>(f : T -> bool, r : Result<E,T>)
    {
      result((res : E) => false,(res : T) => f(res), r)
    }
  }

  module Maybe
  {
    datatype Maybe<T> = Nothing | Just(t : T)

    predicate isJust<T>(m : Maybe<T>)
    {
      match m
      {
      case Nothing => false
      case Just(_) => true
      }
    }

    function maybe<T,A>(def : A, k : T -> A, m : Maybe<T>) : A
    {
      match m {
        case Nothing => def
        case Just(t) => k(t)
      }
    }

    function fmap<T,A>(f : T -> A, m : Maybe<T>) : Maybe<A>
    {
      maybe(Nothing,(res : T) => Just (f(res)),m)
    }

    predicate apply<T>(f : T -> bool, m : Maybe<T>)
    {
      maybe(false,(res : T) => f(res), m)
    }
  }
}
