--- =============================================================================
---
--- A Naive Implementation of Binary Session Types with Choice!
---
--- =============================================================================
|||
||| We extend the naive implementations from `SessionsBinarySimple` by
||| including the idea of `Choice`, that is potential actions that
||| endpoints can take during a protocol's operation.
|||
||| For simplicity, we will *not* realise recursive actions, which
||| require more thought *and* engineering.
|||
||| We will also use the Universe pattern by default.
|||
||| We will not replicate the proofs of duality. These will be part of
||| the *advanced* questions in the exercise sheets.
|||
module SessionsBinaryChoice

||| `Action` captures the communication actions that an endpoint will
||| be required to perform.
|||
||| `Action` has been encoded as a continuation to ensure a normal
||| form and that a series of actions are guaranteed to terminate
||| i.e. Stop!
|||
data Action : Type -> Type where

  ||| Terminate the interaction.
  Stop : Action a

  ||| Select an action to perform, and then continue.
  Select : (options : (List (String, a, Action a)))
                   -> Action a

  ||| Offer a set of options to perform.
  Offer  : (options : (List (String, a, Action a)))
                   -> Action a

-- [ NOTE ]
--
-- We do not offer `Send` nor `Recv` as base terms: Can you explain
-- why we do not, and why `Select` and `Offer` are our only terms
-- (`Stop` aside)?

-- [ Question ]
--
-- Can you think of a way to reduce the number of constructors to
-- just two?

-- [ Question ]
--
-- What is wrong with using `List` here?
-- Can you think of how Vectors can address the problem?

||| Compute the dual of `Actions`.
Dual : Action a -> Action a
Dual Stop
  = Stop

Dual (Select options)
  = Offer (map (\(l,t,a) => (l,t, (Dual a))) options)

Dual (Offer options)
  = Select (map (\(l,t,a) => (l,t, (Dual a))) options)


namespace Example

  data Ty = PING | PONG

  EchoClient : Action Ty
  EchoClient
    = Select [("msg", PING, Offer [("msg", PONG, Stop)])]

  EchoServer : Action Ty
  EchoServer = Dual (EchoClient)


-- [ EOF ]


-- [ Question ]
--
--

-- [ NOTES ]
--
--

-- =============================================================================
--
-- =============================================================================
--
--

{- [ Marker ] ==================================================================

== [ Marker ] =============================================================== -}
