# CS886 Advanced Security-by-Design

The MSc in Cyber Security is divided into three themes that each explore a different
area of cyber security:
Human Factors,
Security-by-Design,
and
Systems Security.

CS886 Advanced Security-by-Design forms the second part of the *Security-by-Design* theme,
which explores more technical elements of cyber security together with a look at the formal modelling and analysis of these technical elements.

The aim of CS886 is to provide students with sufficient grounding and experience over the extent to which formal methods can provide,
at design time,
strong mathematical guarantees about the correctness of our software programs and prevent routinely exploited vulnerabilities from occurring.

Prior to this class, students will have completed the prerequisite courses:

*Security Protocols and Threat Models*
  : which investigates the security of how our code interacts with other external systems;

*Information Security Fundamentals*.
  : a theme agnostic intro course that provides introductory information about security issues affecting code;

CS886 will focus on how we can provide strong guarantees over the security of our code during its design,
using well-studied mathematical techniques and their realisation in tooling.

We will also assume that students are familiar with programming in general and have awareness of functional and imperative styles.

## Getting Started

This repository contains all the public facing material for the course.

`course_information`
:  contains high-level information around the course design including:
   objectives;
   the course syllabus;
   a *rough* plan;
   and
   assessment details;

`etc`
:  files useful for PDF generation using `pandoc`;

`figures`
:  that describe the MSc,
   the Modules,
   and
   the academic week;

`guides`
:  some handy guides on getting Dafny and Idris installed,
   other useful 'howtos' will be placed here;

`handout`
:  documents given to students in advance of the week that details:
   the week's learning objectives;
   the take-home messages;
   and
   the required, recommended, and further reading;

`lecture`
:  LaTeX beamer slides for each week;

`livecode`
:  coding files that will be shown in the second hour of the lecture;

`reading`
:  bibtex files documenting course readings,
   and
   readings organised per week;

`tutorial`
:  document containing the weekly tutorial;

## Important Things to Note

### List 'o Tooling

Tooling used in the design, operation, and realisation of this course:

+ Emacs
+ Org-Mode (Emacs)
+ Pandoc
+ Makefile
+ LaTeX
+ LaTeXmk

### Producing PDFs

All written materials are specified in Markdown, and pandoc enables conversion to other formats.

All PDF materials are generated using LaTeX where suitable, and the LaTeX style files are derived from CIS maintained templates.

The derived versions are found:

https://gitlab.cis.strath.ac.uk/cwb23122/latex

You will need to install some of the folders from this repository into
your local `texmf` tree for optimum results.

That is:

+ `strath` goes to `$HOME/texmf/tex/latex/strath`
+ `branding/strath` goes to `$HOME/texmf/tex/images/strath`

The original templates can be found:

https://gitlab.cis.strath.ac.uk/cis-templates

<!-- [ EOF] -->
